#include "pch.h"
#include "PageImageSource.h"

using namespace concurrency;
using namespace std;
using namespace Microsoft::WRL;
using namespace Windows::UI::Xaml::Media::Imaging;

namespace DxPdfApp
{
	PageImageSource::PageImageSource(Renderer^ renderer, bool fUseVSIS)
		:m_renderer(renderer), m_width(10), m_height(10)
	{
		m_vsis = ref new VirtualSurfaceImageSource(static_cast<int>(m_width), static_cast<int>(m_height), true);
		IInspectable* vsisInspectable = (IInspectable*) reinterpret_cast<IInspectable*>(m_vsis);
		vsisInspectable->QueryInterface(IID_PPV_ARGS(&m_vsisNative));

		ComPtr<IDXGIDevice> dxgiDevice;
		m_renderer->GetDXGIDevice(&dxgiDevice);
		m_vsisNative->SetDevice(dxgiDevice.Get());

		Platform::WeakReference wpThis(this);
		ComPtr<VSISCallBack> spCallBack = Make<VSISCallBack>(wpThis);
		m_vsisNative->RegisterForUpdatesNeeded(spCallBack.Get());

		m_spRenderTask = create_task([] {
			return;
		}, m_cts.get_token());
	}

	PageImageSource::~PageImageSource()
	{
		CancelPageUpdates();
		m_vsisNative.Reset();
	}

	void PageImageSource::SetPage(Windows::Data::Pdf::PdfPage^ pdfPage)
	{
		CancelPageUpdates();
		m_spRenderTask = m_spRenderTask.then([this, pdfPage]() {
			m_pdfPage = pdfPage;
		}, cancellation_token::none());
	}

	void PageImageSource::CancelPageUpdates()
	{
		m_cts.cancel();
		m_cts = cancellation_token_source();
	}

	HRESULT PageImageSource::Resize(double width, double height)
	{
		HRESULT hr = S_OK;
		CancelPageUpdates();
		m_spRenderTask = m_spRenderTask.then([this, width, height]() {
			if (m_width != width || m_height != height)
			{
				m_width = width;
				m_height = height;
				m_vsisNative->Resize(static_cast<int>(m_width), static_cast<int>(m_height));
			}
			RECT updateRect = { 0, 0, static_cast<long>(m_width), static_cast<long>(m_height) };
			m_vsisNative->Invalidate(updateRect);
		}, cancellation_token::none(), task_continuation_context::use_current());
		return hr;
	}


	void PageImageSource::UpdatesNeeded()
	{
		if (m_pdfPage != nullptr)
		{
			ULONG drawingBoundsCount = 0;

			m_vsisNative->GetUpdateRectCount(&drawingBoundsCount);

			std::unique_ptr<RECT[]> drawingBounds(new RECT[drawingBoundsCount]);
			m_vsisNative->GetUpdateRects(drawingBounds.get(), drawingBoundsCount);

			for (ULONG i = 0; i < drawingBoundsCount; ++i)
			{
				RenderPageRect(drawingBounds[i]);
			}
		}
	}

void PageImageSource::RenderPageRect(RECT rect)
{
	m_spRenderTask = m_spRenderTask.then([this, rect]() -> VSISData {
		VSISData vsisData;
		if (!is_task_cancellation_requested())
		{
			HRESULT hr = m_vsisNative->BeginDraw(rect, &(vsisData.dxgiSurface), &(vsisData.offset));
			if (SUCCEEDED(hr))
			{
				vsisData.fContinue = true;
			}
			else
			{
				vsisData.fContinue = false;
			}
		}
		else
		{
			cancel_current_task();
		}
		return vsisData;
	}, m_cts.get_token(), task_continuation_context::use_current()).then([this, rect](task<VSISData> beginDrawTask) -> VSISData {
		VSISData vsisData;
		try
		{
			vsisData = beginDrawTask.get();
			if ((m_pdfPage != nullptr) && vsisData.fContinue)
			{
				ComPtr<IPdfRendererNative> pdfRendererNative;
				m_renderer->GetPdfNativeRenderer(&pdfRendererNative);

				Windows::Foundation::Size pageSize = m_pdfPage->Size;
				float scale = min(static_cast<float>(m_width) / pageSize.Width, static_cast<float>(m_height) / pageSize.Height);

				IUnknown* pdfPageUnknown = (IUnknown*) reinterpret_cast<IUnknown*>(m_pdfPage);
				PDF_RENDER_PARAMS params = PdfRenderParams(D2D1::RectF((rect.left / scale),
					(rect.top / scale),
					(rect.right / scale),
					(rect.bottom / scale)),
					rect.right - rect.left,
					rect.bottom - rect.top,
					D2D1::ColorF(D2D1::ColorF::White),
					FALSE
					);
				/*hr = */pdfRendererNative->RenderPageToSurface(pdfPageUnknown, vsisData.dxgiSurface.Get(), vsisData.offset, &params);
			}
		}
		catch (task_canceled&)
		{

		}
		return vsisData;
	}, cancellation_token::none(), task_continuation_context::use_arbitrary()).then([this](task<VSISData> drawTask) {
		VSISData vsisData;
		try
		{
			vsisData = drawTask.get();
			if (vsisData.fContinue)
				m_vsisNative->EndDraw();
		}
		catch (task_canceled&)
		{

		}
	}, cancellation_token::none(), task_continuation_context::use_current());
}
}
